package de.hawhamburg.reachability;

import org.apache.log4j.Logger;

import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.tools.feeder.SensorDataFeeder;
import de.hawhamburg.reachability.tools.googleCalendar.GoogleCalendarHookService;
import de.hawhamburg.reachability.util.ParameterUtil;

/**
 * A test class which starts a reachability agent. It requires a running ActiveMQ (AMQ) and a MongoDB (MDB).
 *
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR}
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public final class StartTest {

    /**
     * Private constructor to prevent instances of this class may be created.
     */
    private StartTest() { }

    /**
     * Starts a reachability agent.<br><br>
     *
     * <b>Defined command lines arguments:</b><br>
     * <hr>
     * <table>
     *      <tr>
     *          <td><b><i>Flags</i></b><br></td>
     *          <td></td>
     *      </tr>
     *      <tr>
     *          <td><i>--agent</i></td>
     *          <td>Start the reachability agent</td>
     *      </tr>
     *      <tr>
     *          <td><i>--feeder</i></td>
     *          <td>Start the simulation feeder</td>
     *      </tr>
     *      <tr>
     *          <td><i>--calendarService</i></td>
     *          <td>Start the calendar service as well</td>
     *      </tr>
     *      <tr>
     *          <td><i>--info</i></td>
     *          <td>Output of system generated messages of level "INFO"</td>
     *      </tr>
     *      <tr>
     *          <td><i>--debug</i></td>
     *          <td>Output of system generated messages of level "DEBUG"</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td></td>
     *      </tr>
     *      <tr>
     *          <td><b><i>Parameters for simulation feeder</i></b></td>
     *          <td></td>
     *      </tr>
     *      <tr>
     *          <td><i>--storyfile</i></td>
     *          <td>Name of the used story file</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#TEST_FILE})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--storypath</i></td>
     *          <td>Path to the used story file</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#TEST_PATH})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederwait</i></td>
     *          <td>Delay time before start in milliseconds</td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederchannelroot</i></td>
     *          <td>Default AMQ topic prefix</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#TOPIC_ROOT})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederspeed</i></td>
     *          <td>Speed of simulation in simulation minutes per real minut</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#FEEDER_SPEED})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederheartbeat</i></td>
     *          <td>Instruction to send the actual simulation time each simulation minute as time sensor event.</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#FEEDER_HEARTBEAT})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feedersilent</i></td>
     *          <td>Determines if this service shall print out debug output or not.</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#FEEDER_SILENT})</i></td>
     *      </tr>
     * </table>
     *
     * @param args The command line arguments (see method description for usage).
     */
    public static void main(final String[] args) {
        // Remove all existing loggers
        Logger.getRootLogger().removeAllAppenders();
        
        // Local arguments
        String[] localArgs = new String[args.length + 1];
        System.arraycopy(args, 0, localArgs, 0, args.length);
        localArgs[args.length] = "--wrapped";

        // Primary handling of parameters
        String[] subset = {"--agent",
                           "--ui",
                           "--feeder",
                           "--info",
                           "--debug",
                           "--calendarService"};
        ParameterUtil p = ParameterUtil.fromArgs(localArgs, subset);
        boolean none = !p.someset(subset);

        // Overwrite configuration flags
        if (none || p.isset("--info")) {
            MainConfig.setProperty("INFO", true);
        }
        if (none || p.isset("--debug")) {
            MainConfig.setProperty("DEBUG", true);
        }

        // Start requested components
        if (none || p.isset("--agent")) {
            new ReachabilityAgent();
        }
        
        if (p.isset("--feeder")) {
            new SensorDataFeeder(localArgs).start();
        }
        
        if (p.isset("--calendarService")) {
            GoogleCalendarHookService.main(localArgs);
        }
    }
}
