package de.hawhamburg.reachability.properties;

import java.util.Map;
import java.util.Properties;

import de.hawhamburg.reachability.conf.enumeration.ConfigType;

/**
 * Required methods for threads to be property service notifiable.<br><br>
 * 
 * This interfaces defines a list of methods which needs to be implemented by each thread, which would like use a 
 * {@link PropertyService} and receive notifications from it.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public interface PropertyServiceNotifiable {
    
    /**
     * The {@link PropertyService} informs its creator, that there was a change of properties and that the creator
     * should update his properties according to the new ones.
     * 
     * @param newProperties     The list of new {@link Properties}.
     * @param changedProperties The list of changed {@link Properties}.
     * @param propertiesType    The type of properties as {@link ConfigType}.
     */
    void updateProperties(Properties newProperties,
                          Properties changedProperties,
                          ConfigType propertiesType);
    
    /**
     * The {@link PropertyService} requests the actual properties of a certain {@link ConfigType} from its creator. The
     * creator should reload its local property set and return them to the {@link PropertyService}.
     * 
     * @param propertiesType    The {@link ConfigType} for which the properties shall be loaded.
     * @return                  The loaded properties as {@link Properties}.
     */
    Properties loadCurrentProperties(ConfigType propertiesType);

    /**
     * The {@link PropertyService} checks if its creator is running in silent mode, which means the method checks if the
     * component is running in an active process state due to the living place agent life cycle specification.
     * 
     * @return                  <code>true</code> if the component isn't active, otherwise <code>false</code>.
     */
    boolean isSilent();

    /**
     * It may occur that a property is defined as {@link DisplayType#AGENT_DEFINED}, which means that the value range is
     * dynamically defined at runtime. In this case, the {@link PropertyService} asks its creator about the current 
     * value range of this property by using this method.
     * 
     * @param property          The name of the property with dynamic value range as {@link String} for which the 
     *                          current value range is requested.
     * @return                  The actual value range as {@link Map} of {@link String} (value) to {@link String} 
     *                          (display name).
     */
    Map<String, String> getValuesForProperty(String property);
}
