package de.hawhamburg.reachability.ui;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.Timer;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import de.hawhamburg.reachability.StopableObject;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.util.ExternFileUtil;
import de.hawhamburg.reachability.util.OutputUtil;
import de.hawhamburg.reachability.util.XMLUtil;

/**
 * Abstract class for user interface implementation which provides basic methods for simple development<br><br>
 * 
 * This class provides useful methods and functions in common with the develpment of user interfaces for this project.
 * The class provides full {@link ActionListener}, {@link MouseListener}, {@link KeyListener} and {@link WindowListener}
 * support. The required method may need to be implemented within the subclasses.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public abstract class UI extends    JFrame
                         implements ActionListener,
                                    MouseListener,
                                    KeyListener,
                                    StopableObject,
                                    WindowListener {
    

    /**
     * Some unique serial version id.
     */
    private static final long serialVersionUID = 3959878589195891604L;
    
    /**
     * Is this interface started from another interface or for itself?
     */
    private boolean internalStandalone          = false;
    
    /**
     * The windows type.
     */
    private String type                         = null;
    
    /**
     * The parent window, from which this interfaces was created.
     */
    private JFrame parent                       = null;

    /**
     * A random number generator.
     */
    private Random rand                         = new Random();
    
    /**
     * The main {@link JPanel}.
     */
    private JPanel body                         = new JPanel();
    
    /**
     * The repaint {@link Timer}.
     */
    private Timer timer                         = null;
    
    /**
     * The width of the user interface.
     */
    private int windowWidth                     = 0;
    
    /**
     * The height of the user interface.
     */
    private int windowHeight                    = 0;
    
    /**
     * Refresh interval in milliseconds.
     */
    private int refreshInterval                 = 0;
    
    /**
     * The common settings, loaded from the general user interface configuration file.
     */
    private Map<String, String> settingsCommon  = new HashMap<String, String>();
    
    /**
     * The frame settings, loaded from the frame dependent configuration file.
     */
    private Map<String, String> settingsFrame   = new HashMap<String, String>();
    
    /**
     * Protected constructor of this {@link UI} to prevent the creation of direct instances of this class.<br><br>
     * 
     * Creates a new user interface by using {@link #UI(String, JFrame)} with <code>null</code>.
     * 
     * @param title         The title of the user interface as {@link String}.
     */
    protected UI(final String title) {
        this(title, null);
    }
    
    /**
     * Protected constructor of this {@link UI} to prevent the creation of direct instances of this class.<br><br>
     * 
     * Creates a new user interface with the supplied title.
     * 
     * @param title         The title of the user interface as {@link String}.
     * @param parentWindow  The parent window which created this user interface, if existing.
     */
    protected UI(final String title,
                 final JFrame parentWindow) {
        // Overload title to parent constructor
        super(title);
        
        // Set internal class variables
        this.type               = this.getClass().getSimpleName();
        this.parent             = parentWindow;
        this.internalStandalone = this.parent == null;
    }
    
    /**
     * This method builds the interface. It loads the common and local configuration, adds window listeners, positions
     * the window, adds menus, initially builds the interfaces and starts a refresh trigger if required.
     */
    protected final void build() {
        // Load the configuration from external file
        loadConfig();
        
        // Set closing operations and window listener
        setDefaultCloseOperation(DISPOSE_ON_CLOSE);
        addWindowListener(this);
        
        // Center the window
        autoPosition();
        
        // Add (default) menu
        addMenu();
        
        // Try to refresh the first time
        refresh();
        
        // Make the window visible
        setVisible(true);
        
        // Start timer
        if (refreshInterval > 0) {
            this.timer = new Timer(refreshInterval, this);
            this.timer.start();
        }
    }
    
    /**
     * Load the configuration from the default configuration file.
     */
    protected final void loadConfig() {
        
        // Common UI configurations
        Document config = XMLUtil.loadDocument(MainConfig.get("REACHABILITY_UI_CONF_PATH")
                                               + MainConfig.get("REACHABILITY_UI_CONF_FILE"));

        // Common settings
        String xSettings = "/configuration/settings/set";
        List<Element> commonSettings = XMLUtil.elementListByXPath(config, xSettings);
        for (Element set : commonSettings) {
            String key = set.getAttribute("name").trim();
            if (!key.isEmpty()) {
                String value = set.getTextContent().trim();
                settingsCommon.put(key, value);
            }
        }
        
        // Load frame settings
        String xFrame = "/configuration/frames/frame[@id='" + type + "']";
        List<Element> frameSettings = XMLUtil.elementListByXPath(config, xFrame);
        if (frameSettings.size() == 1) {
            Element frameSetting = frameSettings.get(0);
            
            // Refresh timer?
            NodeList refresh = frameSetting.getElementsByTagName("refresh");
            if (refresh.getLength() > 0) {
                try {
                    Element ref = (Element) refresh.item(0);
                    refreshInterval = Integer.parseInt(ref.getAttribute("interval"));
                } catch (NumberFormatException e) {
                    OutputUtil.error("Unable to parse value of attribute 'interval' using '"
                                     + xFrame + "/refresh" + "' on '"
                                     + MainConfig.get("REACHABILITY_UI_CONF_PATH")
                                     + MainConfig.get("REACHABILITY_UI_CONF_FILE"));
                }
            }
            
            // Dimensions
            List<Element> dimensions = null;
            String xDim    = "";
            if (isStandalone()) {
                xDim       = xFrame + "/dimension[@type='standalone']";
                dimensions = XMLUtil.elementListByXPath(config, xDim);
            }

            if (dimensions == null || dimensions.isEmpty()) {
                xDim       =  xFrame + "/dimension[@type='default']";
                dimensions =  XMLUtil.elementListByXPath(config, xDim);
            }
            
            if (dimensions.size() > 0) {
                Element dimension = dimensions.get(0);
                
                // Window height
                try {
                    windowHeight = Integer.parseInt(dimension.getAttribute("height"));
                } catch (NumberFormatException e) {
                    OutputUtil.error("Unable to parse value of attribute 'height' using '"
                                     + xDim + "' on '"
                                     + MainConfig.get("REACHABILITY_UI_CONF_PATH")
                                     + MainConfig.get("REACHABILITY_UI_CONF_FILE"));
                }

                // Window width
                try {
                    windowWidth = Integer.parseInt(dimension.getAttribute("width"));
                } catch (NumberFormatException e) {
                    OutputUtil.error("Unable to parse value of attribute 'width' using '"
                                     + xDim + "' on '"
                                     + MainConfig.get("REACHABILITY_UI_CONF_PATH")
                                     + MainConfig.get("REACHABILITY_UI_CONF_FILE"));
                }
            }
            
            // Local frame settings
            String xFrameSettings = xFrame + "/settings";
            List<Element> settings = XMLUtil.elementListByXPath(config, xFrameSettings);
            if (settings.size() > 0) {
                
                // Get SET elements
                String xFrameSets = xFrameSettings + "[1]/set";
                for (Element set : XMLUtil.elementListByXPath(config, xFrameSets)) {
                    
                    // Get configuration variable name
                    String frameKey = set.getAttribute("name");
                    if (!frameKey.isEmpty()) {
                        
                        // If configuration variable has a valid name, save it's value.
                        String frameValue = set.getTextContent();
                        settingsFrame.put(frameKey, frameValue);
                    }
                }
            }
        }
        
        // Call local configuration if there is a file defined.
        String localFile = MainConfig.get("REACHABILITY_UI_CONF_PATH") + this.getClass().getSimpleName() + ".xml";
        if (ExternFileUtil.isAccessible(localFile)) {
            loadLocalConfig(localFile);
        }
    }
    
    /**
     * Loads the local configuration of the user interface from file. This method needs to be implemented by the
     * sub-class, if required.
     * 
     * @param configFile    The path to the configuration file as {@link String}. 
     */
    protected void loadLocalConfig(final String configFile) { }
    
    /**
     * The method centers the user interface to the middle of the screen.
     */
    protected final void autoPosition() {
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        
        int posX, posY;
        if (this.parent == null) {
            // Screen Center
            posX = (screenSize.width - windowWidth) / 2;
            posY = (screenSize.height - windowHeight) / 2;
        } else {
            // Same position like parent window
            Rectangle parentPosition = parent.getBounds();
            posX = parentPosition.x;
            posY = parentPosition.y;
        }
        
        setBounds(posX, posY, windowWidth, windowHeight);
    }
    
    /**
     * Corrects the position and size of the window by {@link #autoPosition()} if <i>refresh</i> is supplied as
     * <code>true</code>, otherwise only the size will be updated.
     * 
     * @param refresh       If set to <code>true</code> only the size will be updated, otherwise position as well.
     */
    protected final void fixBounds(final boolean refresh) {
        if (refresh) {
            Rectangle bounds = this.getBounds();
            setBounds(bounds.x, bounds.y, windowWidth, windowHeight);
        } else {
            autoPosition();
        }
    }
    
    /**
     * This method adds a menu to the user interface. This method needs to be implemented by the sub-class, if a menu is
     * required.
     */
    protected abstract void addMenu();
    
    /**
     * Returns a user interface grid.
     * 
     * @param rowsIn        Amount of rows to be displayed
     * @param colsIn        Amount of columns to be displayed
     * @return              Returns a user interface grid.
     */
    protected final GridLayout getGrid(final int rowsIn,
                                       final int colsIn) {
        return new GridLayout(rowsIn, colsIn);
    }
    
    /**
     * Returns the body of the the user interface as {@link JPanel}.
     * 
     * @return              The interfaces body as {@link JPanel}.
     */
    protected final JPanel getBody() {
        return this.body;
    }

    /**
     * Returns the parent window if existing as {@link JFrame}, otherwise <code>null</code>.
     * 
     * @return              The parent window as {@link JFrame} if existing, otherwise <code>null</code>.
     */
    protected final JFrame getParentWindow() {
        return this.parent;
    }

    /**
     * Returns the timer thread as {@link Timer} if existing, otherwise <code>null</code>.
     * 
     * @return              The timer thread as {@link Timer} if existing, otherwise <code>null</code>.
     */
    protected final Timer getTimer() {
        return this.timer;
    }

    /**
     * Returns the height of this window as integer.
     * 
     * @return              The height of this window as integer.
     */
    protected final int getWindowHeight() {
        return this.windowHeight;
    }

    /**
     * Sets/changes the height of this window to a new value.
     * 
     * @param height        The new height of this window as integer.
     */
    protected final void setWindowHeight(final int height) {
        this.windowHeight = height;
    }

    /**
     * Returns the width of this window as integer.
     * 
     * @return              The width of this window as integer.
     */
    protected final int getWindowWidth() {
        return this.windowWidth;
    }

    /**
     * Sets/changes the width of this window to a new value.
     * 
     * @param width         The new width of this window as integer.
     */
    protected final void setWindowWidth(final int width) {
        this.windowWidth = width;
    }
    
    /**
     * Returns the value of a configuration variable specified by <i>key</i> as {@link String} if existing, otherwise 
     * <code>null</code>.
     * 
     * @param key           The name of the configuration variable as {@link String}.
     * @return              The value of this configuration variable as {@link String} if existing, otherwise 
     *                      <code>null</code>.
     */
    protected final String getConfigValue(final String key) {
        String result = getFrameConfigValue(key, true);
        if (result == null) {
            result = getCommonConfigValue(key, true);
        }
        if (result == null) {
            OutputUtil.error("[getConfigValue] Requested setting '" + key + "' was not defined in '"
                             + MainConfig.get("REACHABILITY_UI_CONF_PATH")
                             + MainConfig.get("REACHABILITY_UI_CONF_FILE")
                             + "'. NULL will be returned instead.");
        }
        return result;
    }
    
    /**
     * Returns the value of a configuration variable specified by <i>key</i> as {@link Integer} if existing, otherwise
     * <code>null</code>.
     * 
     * @param key           The name of the configuration variable as {@link String}.
     * @return              The value of this configuration variable as {@link Integer} if existing, otherwise 
     *                      <code>null</code>.
     */
    protected final Integer getConfigValueAsInteger(final String key) {
        Integer result = null;
        try {
            result = Integer.valueOf(getConfigValue(key));
        } catch (NumberFormatException e) {
            OutputUtil.error("[getConfigValueAsInteger][NumberFormatException] " + e.getMessage());
        }
        return result;
    }
    
    /**
     * Returns a common configuration value as {@link String} specified by the <i>key</i> if existing, otherwise 
     * <code>null</code>.
     * 
     * @param key           The name of the common configuration variable as {@link String}.
     * @return              The value of this common configuration variable as {@link String} if existing, otherwise 
     *                      <code>null</code>.
     */
    protected final String getCommonConfigValue(final String key) {
        return getCommonConfigValue(key, false);
    }
    
    /**
     * Returns a common configuration value as {@link String} specified by the <i>key</i> if existing, otherwise 
     * <code>null</code>.
     * 
     * @param key           The name of the common configuration variable as {@link String}.
     * @param internalCall  If <code>true</code> and the variable couldn't be found, no error will be published.
     * @return              The value of this common configuration variable as {@link String} if existing, otherwise 
     *                      <code>null</code>.
     */
    private String getCommonConfigValue(final String key,
                                        final boolean internalCall) {
        if (!settingsCommon.containsKey(key)) {
            if (!internalCall) {
                OutputUtil.error("[getCommonConfigValue] Requested common setting '" + key
                                 + "' was not defined in '"
                                 + MainConfig.get("REACHABILITY_UI_CONF_PATH")
                                 + MainConfig.get("REACHABILITY_UI_CONF_FILE")
                                 + "'. NULL will be returned instead.");
            }
            return null;
        }
        return settingsCommon.get(key);
    }
    
    /**
     * Returns a frame (local) configuration value as {@link String} specified by the <i>key</i> if existing, otherwise 
     * <code>null</code>.
     * 
     * @param key           The name of the frame (local) configuration variable as {@link String}.
     * @return              The value of this the frame (local) configuration variable as {@link String} if existing, 
     *                      otherwise <code>null</code>.
     */
    protected final String getFrameConfigValue(final String key) {
        return getFrameConfigValue(key, false);
    }
    
    /**
     * Returns a frame (local) configuration value as {@link String} specified by the <i>key</i> if existing, otherwise 
     * <code>null</code>.
     * 
     * @param key           The name of the frame (local) configuration variable as {@link String}.
     * @param internalCall  If <code>true</code> and the variable couldn't be found, no error will be published.
     * @return              The value of this frame (local) configuration variable as {@link String} if existing, 
     *                      otherwise <code>null</code>.
     */
    private String getFrameConfigValue(final String key,
                                       final boolean internalCall) {
        if (!settingsFrame.containsKey(key)) {
            if (!internalCall) {
                OutputUtil.error("[getFrameConfigValue] Requested frame setting '" + key + "' was "
                                 + "not defined in '"
                                 + MainConfig.get("REACHABILITY_UI_CONF_PATH")
                                 + MainConfig.get("REACHABILITY_UI_CONF_FILE")
                                 + "'. NULL will be returned "
                                 + "instead.");
            }
            return null;
        }
        return settingsFrame.get(key);
    }
    
    /**
     * Returns the value of a frame (local) configuration variable specified by <i>key</i> as {@link Integer} if 
     * existing, otherwise <code>null</code>.
     * 
     * @param key           The name of the frame (local) configuration variable as {@link String}.
     * @return              The value of this frame (local) configuration variable as {@link Integer} if existing, 
     *                      otherwise <code>null</code>.
     */
    protected final Integer getFrameConfigValueAsInteger(final String key) {
        Integer result = null;
        try {
            result = Integer.valueOf(getFrameConfigValue(key));
        } catch (NumberFormatException e) {
            OutputUtil.error("[getFrameConfigValueAsInteger][NumberFormatException] " + e.getMessage());
        }
        return result;
    }
    
    /**
     * Adds an element as {@link Component} to the {@link #body} of this user interface.
     * 
     * @param x             The {@link Component} which shall be added to the {@link #body} of this interface.
     */
    protected final void addBody(final Component x) {
        this.body.add(x);
    }
    
    /**
     * Determines if the interface is standing alone (has no parent window).
     * 
     * @return              <code>true</code> if its standing alone (has no parent window), otherwise 
     *                      <code>false</code>.
     */
    public final boolean isStandalone() {
        return this.internalStandalone;
    }
    
    /**
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param e             An {@link ActionEvent}.
     */
    public void actionPerformed(final ActionEvent e) { }

    /**
     * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param e             A {@link MouseEvent}.
     */
    public void mouseClicked(final MouseEvent e) { }

    /**
     * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param e             A {@link MouseEvent}.
     */
    public void mouseEntered(final MouseEvent e) { }

    /**
     * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param e             A {@link MouseEvent}.
     */
    public void mouseExited(final MouseEvent e) { }

    /**
     * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param e             A {@link MouseEvent}.
     */
    public void mousePressed(final MouseEvent e) { }

    /**
     * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param e             A {@link MouseEvent}.
     */
    public void mouseReleased(final MouseEvent e) { }

    /**
     * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param arg0          A {@link KeyEvent}.
     */
    public void keyPressed(final KeyEvent arg0) { }

    /**
     * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param arg0          A {@link KeyEvent}.
     */
    public void keyReleased(final KeyEvent arg0) { }

    /**
     * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param arg0          A {@link KeyEvent}.
     */
    public void keyTyped(final KeyEvent arg0) { }
    
    
    /**
     * @see java.awt.event.WindowListener#windowActivated(java.awt.event.WindowEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param arg0          A {@link WindowEvent}.
     */
    public void windowActivated(final WindowEvent arg0) { }

    /**
     * @see java.awt.event.WindowListener#windowClosed(java.awt.event.WindowEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param arg0          A {@link WindowEvent}.
     */
    public void windowClosed(final WindowEvent arg0) { }

    /**
     * @see java.awt.event.WindowListener#windowClosing(java.awt.event.WindowEvent)
     * 
     * @param arg0          A {@link WindowEvent}.
     */
    public final void windowClosing(final WindowEvent arg0) {
        this.stopTimer();
        this.shutDown();
    }

    /**
     * @see java.awt.event.WindowListener#windowDeactivated(java.awt.event.WindowEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param arg0          A {@link WindowEvent}.
     */
    public void windowDeactivated(final WindowEvent arg0) { }

    /**
     * @see java.awt.event.WindowListener#windowDeiconified(java.awt.event.WindowEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param arg0          A {@link WindowEvent}.
     */
    public void windowDeiconified(final WindowEvent arg0) { }

    /**
     * @see java.awt.event.WindowListener#windowIconified(java.awt.event.WindowEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param arg0          A {@link WindowEvent}.
     */
    public void windowIconified(final WindowEvent arg0) { }

    /**
     * @see java.awt.event.WindowListener#windowOpened(java.awt.event.WindowEvent)
     * 
     * This method needs to be specified by the sub-class, if required.
     * 
     * @param arg0          A {@link WindowEvent}.
     */
    public void windowOpened(final WindowEvent arg0) { }

    /**
     * Destroys this interface. 
     */
    public final void killWindow() {
        this.dispatchEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSING));
    }
    
    /**
     * Stop a running refresh timer if existing.
     */
    protected final void stopTimer() {
        if (this.timer != null) {
            this.timer.stop();
            this.timer = null;
        }
    }
    
    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     * 
     * This method needs to be specified by the sub-class, if required.
     */
    public abstract void shutDown();
    
    /**
     * Creates an information dialog containing a specified message.
     * 
     * @param msg           A message as {@link String}.
     */
    protected final void infoDialog(final String msg) {
        JOptionPane.showMessageDialog(this, msg);
    }

    /**
     * Builds the {@link UI} by using {@link #refresh(boolean)} with <code>false</code>.
     */
    public final void refresh() {
        refresh(false);
    }
    
    /**
     * (Re-)builds the rendering of the {@link UI}. This method needs to be specified by the sub-class.
     * 
     * @param refresh       <code>true</code> indicates that this method is called with the intention to refresh the 
     *                      page with new data, <code>false</code> means that it was called as initial build.
     */
    public abstract void refresh(final boolean refresh);
    
    /**
     * Generates a random ID for this {@link UI}.
     * 
     * @return              Returns a randomly generated identity number.
     */
    public final String getId() {
        return MainConfig.get("REACHABILITY_UI_NAME") + "_" + type + "_" + rand.nextInt(Integer.MAX_VALUE);
    }
}
